local mob = Class()

local function dom(val) return val < 0.5 end 

local function split(a, b, threshold)
    return random() < (threshold or 0.5) and a or b
end

local function split_mutate(a, b, threshold, mutationChance, mutationMagnitude)
    local gene = random() < (threshold or 0.5) and a or b
    if random() < (mutationChance or 0.5) then gene = gene + random(-mutationMagnitude, mutationMagnitude) end

    if gene < 0 then gene = 0 elseif gene > 1 then gene = 1 end

    return gene
end

local function average(a, b)
    return (a + b) / 2
end

local function average_mutate(a, b, mutationChance, mutationMagnitude)
    local average = (a + b) / 2 

    if random() < mutationChance then
        average = average + random(-mutationMagnitude, mutationMagnitude)
    end
    if average < 0 then average = 0 elseif average > 1 then average = 1 end

    return average
end

function mob:init(x, y, i)
    self.position = Vector(x or CENTER.x, y or CENTER.y)
    self.angle = math.pi * 2 * random()

    self.genes = {}

    self.dead = false
    self.social = 1
    self.life = 1
    self.fertile = false

    self.lastPartner = false

    -- r, g, b, rad, speed, left, right, turn, lonelyrate, deathrate
    for index=1, 11 do
        self.genes[index] = 0.5
    end
end

function mob:breed(b, x, y)
    local child = Mob(x or self.position.x, y or self.position.y) 
    -- child.lastPartner = b
    for i,v in ipairs(child.genes) do
        child.genes[i] = split_mutate(self.genes[i], b.genes[i], 0.5, 0.2, 0.01)
        -- local function split_mutate(a, b, threshold, mutationChance, mutationMagnitude)
    end

    return child
end

function mob:diff(b)
    local difference = 0
    for i=1, #self.genes do
        local a, b = self.genes[i], b.genes[i] 
        difference = difference + (a - b)
    end
    return difference
end

function mob:geneSum()
    local sum = 0
    for i,v in ipairs(self.genes) do
        sum = sum + v
    end
    return sum
end

function mob:dieALittle(m)
    self.life = self.life - (self.genes[10] * 0.01 * (m or 1))
    if self.life < 0.2 then self.fertile = true end
    if self.life < 0 then self.dead = true end
end

function mob:update(dt)
    self:dieALittle()
    if random() < self.genes[6] then
        self.angle = self.angle + self.genes[8] * 0.5
    end

    if random() < self.genes[7] then
        self.angle = self.angle - self.genes[8] * 0.5
    end

    local nx, ny = self.position.x + self.genes[5] * cos(self.angle), self.position.y + self.genes[5] * sin(self.angle)


    if nx < 0 then nx = WIDTH elseif nx > WIDTH then nx = 0 end
    if ny < 0 then ny = HEIGHT elseif ny > HEIGHT then ny = 0 end

    self.position:set(nx, ny)
    MOBS:update(self, nx, ny)
end

function mob:draw()
    lg.setColor(self.genes[1], self.genes[2], self.genes[3])

    -- local mode = "fill"
    -- if dom(self.genes[5]) then mode = "line" end
    lg.circle("fill", self.position.x, self.position.y, scale)

    lg.setColor(self.genes[1], self.genes[2], self.genes[3], 0.2)
    lg.circle("line", self.position.x, self.position.y, breedRadius * self.genes[11])

    -- lg.setColor(Color.white)
    -- lg.print(floor(self:geneSum() * 100) / 100, self.position.x, self.position.y)

    -- if self.lastPartner then
    --     lg.line(self.position.x, self.position.y, self.lastPartner.position.x, self.lastPartner.position.y)
    -- end
end

return mob
