
--:: SHORTHANDS ::--
lg = love.graphics
fs = love.filesystem
kb = love.keyboard
lm = love.mouse
lp = love.physics
random = love.math.random
noise = love.math.noise
sin = math.sin
cos = math.cos
floor = math.floor
ceil = math.floor
f = string.format

-- Modifying lg.setColor
local setColor = lg.setColor
love.graphics.setColor = function(r, g, b, a)
    if type(r) == "number" and not g and not b and not a then
        setColor(r, r, r, 1)
    else
        setColor(r, g, b, a)
    end
end

-- Modifting lua print
local _print = print
function print(...) _print(string.format("%s:%d:", debug.getinfo(2).short_src, debug.getinfo(2).currentline), ...) end

function love.load()
    -- Loading util first as it contains the "requireDirectory" function used to load everything else
    Util = require "lib.Util"
    Util.requireDirectory("lib")

    -- GLOBALS
    WIDTH = lg.getWidth()
    HEIGHT = lg.getHeight()
    CENTER = Vector(WIDTH / 2, HEIGHT / 2)

    Util.requireDirectory("class")

    -- Keypress events
    Input:keypress("escape", love.event.push, "quit")
    Input:keypress("f1", love.system.openURL, "file://"..love.filesystem.getSaveDirectory())
    Input:keypress("f2", love.graphics.captureScreenshot, os.time() .. ".png")
    Input:keypress("space", function() 
        if not started then started = true end
        run = not run 
    end)

    lg.setDefaultFilter("nearest", "nearest")
    lg.setBackgroundColor(Color.darkgray)

    font = {
        large = lg.newFont(24),
        small = lg.newFont(16)
    }

    initPopulation = 400
    maxPopulation = 800
    breedRadius = 32
    socialRadius = 64
    breedChance = 0.1
    crowdedThresh = 20

    tick = 0
    rate = 60
    time = 0
    stepsPerTick = 1
    run = false
    started = false

    spawn()


    drawnMobs = 0

    scale = 1

    canvas = lg.newCanvas(WIDTH / scale, HEIGHT / scale)
end

function spawn(amt)
    MOBS = spatial()

    for i=1, (amt or initPopulation) do
        local x, y = random(WIDTH), random(HEIGHT)
        MOBS:insert(Mob(x, y), x, y)
    end
end

function step()
    if not run then return end
    local steps = stepsPerTick
    if kb.isDown("s") then steps = stepsPerTick * 10 end
    for i=1, steps do
        local list, len = MOBS:queryRect(0, 0, WIDTH, HEIGHT)

        for _, mob in ipairs(list) do
            local x, y = mob.position.x - (socialRadius / 2), mob.position.y - (socialRadius / 2)
            local friends, friendCount = MOBS:queryRect(x, y, socialRadius / 2, socialRadius / 2)

            -- Lonely stuff
            if friendCount < 1 then
                mob:dieALittle(4)
            elseif friendCount > crowdedThresh then
                mob:dieALittle(4)
            else
                mob.social = mob.social + 0.01
                if mob.social > 1 then mob.social = 1 end

                -- Fucking
                for _, mate in ipairs(friends) do
                    if mob ~= mate then -- cant fuck yourself right
                        if Maf.distance(mob.position.x, mob.position.y, mate.position.x, mate.position.y) < breedRadius * mob.genes[11] then
                            if mob.fertile and mate.fertile then
                                if random() < breedChance and MOBS.length < maxPopulation then
                                    local x = (mob.position.x + mate.position.x) / 2
                                    local y = (mob.position.y + mate.position.y) / 2
                                    MOBS:insert(mob:breed(mate, x, y), x, y)
                                end
                            end
                        end
                    end
                end
            end

            mob:update()
            if mob.dead then
                MOBS:remove(mob)
            end
        end
    end
end

function love.update(dt)
    Input:update()

    tick = tick + dt
    if tick > (1 / rate) then
        step()
        tick = 0
    end
end

function love.draw()

    local list, len = MOBS:queryRect(0, 0, WIDTH, HEIGHT)
    drawnMobs = len

    lg.setColor(Color.darkgray)
    lg.rectangle("fill", 0, 0, WIDTH, HEIGHT)

    lg.setCanvas(canvas)
    lg.clear()
    -- lg.setBlendMode("add")
    lg.push()
    lg.scale(1 / scale, 1 / scale)
    for _, mob in ipairs(list) do
        mob:draw()
    end

    lg.pop()

    -- lg.setBlendMode("alpha")
    -- lg.setColor(0, 0, 0, 0.01)
    -- lg.rectangle("fill", 0, 0, WIDTH, HEIGHT)
    lg.setCanvas()

    lg.setColor(Color.white)
    lg.draw(canvas, 0, 0, 0, scale, scale)


    -- Teckst
    lg.setColor(Color.red)
    lg.setFont(font.small)
    lg.printf("FPS: " .. love.timer.getFPS() .. "\nPopulation: " .. MOBS.length, 12, 12, WIDTH, "left")
    --
    local text = [[

    ]]
    if not started then
        lg.setColor(Color.offwhite)
        lg.setFont(font.small)
        lg.printf(text, 0, 100, WIDTH, "center")
    end
end

function hud()

end


function love.keypressed(key)
    Input:keypressed(key)
end

function love.keyreleased(key) 
    Input:keyreleased(key)
end

function love.mousepressed(x, y, key) 
    Input:mousepressed(x, y, key)
end

function love.mousereleased(x, y, key) 
    Input:mousereleased(x, y, key)
end
